% FILE:        gem1.m
% PURPOSE:     Machine to execute self-describing grammars
%              Robust version
% EXAMPLE:     
%   obj = gem1();                         % constructor
%   dbg = obj.nowhite;                    % a deblanking grammar
%   o = obj.run('x y z', dbg)             % deblank string
%   o = obj.run('x y', dbg, 'T')          % .. and trace steps
% REQUIRES:
%   iog1.c iog1.h
% TEST:
%   testGem1()

% COPYRIGHT:   1988 W. M. McKeeman.  See license.txt.
% MODS:	       1988.06.02 -- mckeeman -- original
%              2008.06.11 -- mckeeman@dartmouth -- moved to MEX
%              2009.04.08 -- luhe@dartmouth     -- added classes and nowhite

function obj = gem1()
  EOL = char(10);                                % ASCII end of line
  obj = public();
  %return;
  
  % ----------------------------------------------------
  
  % Run GEM.  gtxt must be deblanked.
  function otxt = run(itxt, gtxt, flag)
    if nargin == 2, flag = ''; end
    otxt = iog1(itxt, gtxt, flag);
  end

  function o = public()
    o = struct;
    o.run  = @run;  % execute GEM

    %    ------- input grammars -----------------
    
    % digits   (defines phrase name d) 
    o.digitCFG = sprintf('d=''%c'';','0':'9');
    % lower case letters  (defines phrase name l)
    o.lowerCFG = sprintf('l=''%c'';','a':'z');
    % upper case letters  (defines phrase name l)
    o.upperCFG = sprintf('l=''%c'';','A':'Z');
    % ASCII  (defines phrase name a)
    o.asciiCFG = sprintf('a=''%c'';',char(1:126));            
            
    % -------------- input/output grammars ----------------       
    % digits   (defines phrase name D)
    o.digitIOG = sprintf('D=''%c''"%c";',repeat('0':'9', 2));
    % lower case letters  (defines phrase name L)
    o.lowerIOG = sprintf('L=''%c''"%c";',repeat('a':'z', 2));
    % upper case letters  (defines phrase name L)
    o.upperIOG = sprintf('L=''%c''"%c";',repeat('A':'Z', 2));
    % asciiIO  (defines phrase names A)
    o.asciiIOG = sprintf('A=''%c''"%c";', repeat(char(1:126), 2));
    % !"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~ 
    
    % A prescanned scanner (no whitespace)
    o.nowhite = [
      'g=pg;'          ...   list of p-s
      'g=;'            ...
      'p='' '';'       ...   discard blank
      'p=''' EOL ''';' ...   discard EOL
      'p=IAI;'         ...   keep input symbols
      'p=OAO;'         ...   keep output symbols
      'p=A;'           ...   keep everything else
      'I=''''''"''";'  ...   see ', send '
      'O=''"''""";'    ...   see ", send "
      o.asciiIOG
      ];
  
  
    % A pretty printer
    o.pretty = [
      'g = r g;'                       EOL ...  list of r-s
      'g =;'                           EOL ...
     ['r = L ''='' " " "=" f '';'' ";" "' EOL '";'] EOL ...
      'f = " " p f;'                   EOL ...
      'f =;'                           EOL ...
      'p = I A I;'                     EOL ...  pass input symbols
      'p = O A O;'                     EOL ...  pass output symbols
      'p = L;'                         EOL ...  pass letters
      'I = '''''' "''";'               EOL ...  see ', send '
      'O = ''"'' """;'                 EOL ...  see ", send "
      o.lowerIOG                       EOL ...  a:z
      o.upperIOG                       EOL ...  A:Z
      o.asciiIOG
      ];
    
    o.invert = [
      'g = p g;'                       EOL ...
      'g = ;'                          EOL ...
      'p = '''''' """ A '''''' """;'   EOL ...
      'p = ''"'' "''" A ''"'' "''";'   EOL ...
      'p = A;'                         EOL ...
      o.asciiIOG
      ];
    % a right associative grammar for + and -
    o.sum    = [
      'g = e         "0";'             EOL ...
      'e = t ''+'' e   "1";'           EOL ...
      'e = t ''-'' e   "2";'           EOL ...
      'e = t         "3";'             EOL ...
      't = ''x''       "4";'           EOL ...
      ];
    
  end

end
