% FILE:        gem3.m
% PURPOSE:     Machine to execute self-describing grammars
%              Built-in character classes
% EXAMPLE:     
%   obj = gem3();                        % constructor
%   gr  = 'g = dg; g = d;';              % digit strings 
%   o = obj.run('123', gr, 'd');         % use digitCFG built-in
%   isempty(o);                          % recognizes '123'
% REQUIRES:
%   iog3.c iog3.h
% TEST:
%   testGem3()

% COPYRIGHT:   1988 W. M. McKeeman.  See license.txt.
% MODS:	    1988.06.02 -- mckeeman -- original
%           2008.06.11 -- mckeeman@dartmouth -- moved to MEX
%           2009.04.09 -- luhe@dartmouth     -- use built-in classes

function obj = gem3()
  EOL = char(10);                                % ASCII end of line
  obj = public();
  %return;
  
  % ----------------------------------------------------
  
  % Run GEM with builtin character classes
  function otxt = run(itxt, gtxt, flags)
    if nargin == 2, flags = ''; end
    otxt = iog3(itxt, gtxt, flags);              % call mex with flags
  end

  function o = public()
    o = struct;
    o.run  = @run;  % execute GEM
    
    % single-up multi-character Vi and Vo elements
    o.nomulti = [   % needs flag A
      'g=P;'            ...  list of p-s
      'p=III;'          ...  keep '''
      'p=OOO;'          ...  keep """
      'p=''''''i;'      ...  expand 'input stuff'
      'p=''"''o;'       ...  expand "output stuff"
      'p=A;'            ...  pass everything else
      'i='''''';'       ...  comes after input stuff
      'i="''"A"''"i;'   ...  expand one  input char
      'o=''"'';'        ...  comes after output stuff
      'o="""A"""o;'     ...  expand one output char
      'I=''''''"''";'   ...  see ', send '
      'O=''"''""";'     ...  see ", send "
      'P=pP;'           ...  list of p-s
      'P=;'             ...
    ];

    % A pretty printer -- remove excess and insert whitespace
    o.pretty = [   % needs flag ULA
      'g = B R;'                       EOL ...
     ['r = L B ''='' " " "=" F B '';'' B ";" "' EOL '";'] EOL ...
      'f = B " " p;'                   EOL ...
      'p = I I I;'                     EOL ...  single quote
      'p = I i "''";'                  EOL ...  pass input symbols
      'p = O O O;'                     EOL ...  pass output symbols
      'p = O o """;'                   EOL ...
      'p = L;'                         EOL ...  pass letters
      'i = '''''';'                    EOL ...  end of input symbol
      'i = A i;'                       EOL ...
      'o = ''"'';'                     EOL ...  end of output symbol
      'o = A o;'                       EOL ...
      'I = '''''' "''";'               EOL ...  see ', send '
      'O = ''"'' """;'                 EOL ...  see ", send "
      'R = r R;'                       EOL ...  list of r-s
      'R =;'                           EOL ...
      'F = f F;'                       EOL ...  list of f-s
      'F =;'                           EOL ...
      'B = b B;'                       EOL ...  list of b-s
      'B =;'                           EOL ...
      'b = '' '';'                     EOL ...  literal blank
     ['b = ''' EOL ''';']              EOL ...  literal newline
      ];
    
    o.invert = [   % needs flag A
      'g = p g;'                       EOL ...
      'g = ;'                          EOL ...
      'p = '''''' """ A '''''' """;'   EOL ...
      'p = ''"'' "''" A ''"'' "''";'   EOL ...
      'p = A;'                         EOL ...
      ];

    % making expression postfix code
    o.postfix   = [  % needs flag LD
      'g = e;'                         EOL ...
      'e = t R;'                       EOL ...
      'r = ''+'' t "+";'               EOL ...
      'r = ''-'' t "-";'               EOL ...
      't = f S;'                       EOL ...
      's = ''*'' f "*";'               EOL ...
      's = ''/'' f "/";'               EOL ...
      'f = L;'                         EOL ...
      'f = D;'                         EOL ...
      'f = ''('' e '')'';'             EOL ...
      'R = r R;'                       EOL ...
      'R =;'                           EOL ...
      'S = s S;'                       EOL ...
      'S =;'                           EOL ...
      ];

    % making X86 type double code
    o.x86   = [        % needs flag LD
      'g = e;'                         EOL ...
      'e = t R;'                       EOL ...
     ['r = ''+'' t "fadd' EOL '";']    EOL ...
     ['r = ''-'' t "fsub' EOL '";']    EOL ...
      't = f S;'                       EOL ...
     ['s = ''*'' f "fmul' EOL '";']    EOL ...
     ['s = ''/'' f "fdiv' EOL '";']    EOL ...
     ['f = "fld " L "' EOL '";']       EOL ...
     ['f = "fld =" D "' EOL '";']      EOL ...
      'f = ''('' e '')'';'             EOL ...
      'R = r R;'                       EOL ...
      'R =;'                           EOL ...
      'S = s S;'                       EOL ...
      'S =;'                           EOL ...
     ];
  
     % left associative arithmetic expression to operator-early PFN
    o.prefix  = [      % needs flag LD
      'g = e;'                         EOL ...
      'e = "+" t ''+'' e;'             EOL ...
      'e = "-" t ''-'' e;'             EOL ...
      'e = t;'                         EOL ...
      't = "*" f ''*'' t;'             EOL ...
      't = "/" f ''/'' t;'             EOL ...
      't = f;'                         EOL ...
      'f = L;'                         EOL ...
      'f = D;'                         EOL ...
      'f = ''('' e '')'';'             EOL ...
      ];

  end

end
