% FILE:    gem4.m
% PURPOSE: machine to execute self-describing grammars
% METHOD:  The machine iog4.c has builtin multi-character
%   input and output symbols.
% EXAMPLES:     
%   G = gem4();                          % constructor

% COPYRIGHT:   1988 W. M. McKeeman.  See license.txt.
% MODS:	       1988.06.02 -- mckeeman -- original
%              2008.06.11 -- mckeeman@dartmouth -- moved to MEX

function obj = gem4()
  EOL = char(10);                        % ascii end-of-line
  
  %    ------- input grammars -----------------
  
  % de-white-er
  nowhite = [        % requires A
    'g=pg;'           ...
    'g=;'             ...
    'p='' '';'        ...  discard blank
    'p=''' EOL ''';'  ...  discard EOL
    'p=III;'          ...  keep '''
    'p=OOO;'          ...  keep """
    'p=''''''i;'      ...  expand 'input stuff'
    'p=''"''o;'       ...  expand "output stuff"
    'p=A;'            ...  pass everything else
    'i='''''';'       ...  comes after input stuff
    'i="''"A"''"i;'   ...  expand one  input char
    'o=''"'';'        ...  comes after output stuff
    'o="""A"""o;'     ...  expand one output char
    'I=''''''"''";'   ...  see ', send '
    'O=''"''""";'     ...  see ", send "
    ];
  
  % replace a* with A in a grammar (only lower) 
  % a* goes to A
  nostar1 = [        % requires ALU
    'g = r g;'                               EOL ...
    'g =;'                                   EOL ...
    'r = L ''='' "="  f '';'' ";" ;'         EOL ...
    'f = p f;'                               EOL ...
    'f =;'                                   EOL ...
    'p = I A I;'                             EOL ...
    'p = O A O;'                             EOL ...
    'p = s;'                                 EOL ...
    'p = L;'                                 EOL ...
    'I = '''''' "''";'                       EOL ...  see ', send '
    'O = ''"''  """;'                        EOL ...  see ", send "
    sprintf('s=''%c*''"%c";', shuffle('a':'z','A':'Z'))  EOL ...
    ];

  % make new rules to define a* (only lower)     
  % 'a*' goes to "A=aA;A=;";
  % nostar2 can make multiple copies of a recursive rule.
  % Logically that is OK; it is awful for performance.
  % So, we use nodup (below) to clean out the duplicates
  nostar2 = [         % requires ALU   
    'g = r g;'                               EOL ...
    'g =;'                                   EOL ...
    'r = l ''='' f '';'';'                   EOL ...
    'f = p f;'                               EOL ...
    'f =;'                                   EOL ...
    'p = '''''' a '''''';'                   EOL ...
    'p = ''"'' a ''"'';'                     EOL ...
    'p = s;'                                 EOL ...
    'p = l;'                                 EOL ...
    sprintf(sprintf('s=''%c*''"%%c=%c%%c;%%c=;";', ...
      repeat('a':'z',2)),repeat('A':'Z',3))  EOL ...
    ];

  % replace a+ in a grammar (only lower) 
  % a+ goes to aa*
  noplus = [         % requires ALU
    'g = r g;'                               EOL ...
    'g =;'                                   EOL ...
    'r = L ''='' "="  f '';'' ";" ;'         EOL ...
    'f = p f;'                               EOL ...
    'f =;'                                   EOL ...
    'p = I A I;'                             EOL ...
    'p = O A O;'                             EOL ...
    'p = S;'                                 EOL ...
    'p = L;'                                 EOL ...
    'p = L ''*'' "*";'                       EOL ...
    'I = '''''' "''";'                       EOL ...  see ', send '
    'O = ''"''  """;'                        EOL ...  see ", send "
    sprintf('S=''%c+''"%c%c*";', repeat('a':'z', 3))  EOL ...
    ];
  
  % an IOG pretty printer
  % handles extra and missing whitespace, * and +
  % leaves multicharacter inputs and outputs unchanged
  pretty0 = [        % requires ALU
    'g = b* r*;'                             EOL ...
   ['r = L b* ''='' " =" f* b* '';'' b* ";' EOL '";'] EOL ...
    'f = b* " " p;'                          EOL ...
    'p = I I I;'                             EOL ... 
    'p = I i "''";'                          EOL ...
    'p = O O O;'                             EOL ...
    'p = O o """;'                           EOL ...
    'p = L b* ''*'' "*";'                    EOL ...
    'p = L b* ''+'' "+";'                    EOL ...
    'p = L;'                                 EOL ...
    'i = '''''';'                            EOL ...
    'i = A i;'                               EOL ...
    'o = ''"'';'                             EOL ...
    'o = A o;'                               EOL ...
    'I = '''''' "''";'                       EOL ...
    'O = ''"'' """;'                         EOL ...
    'b = '' '';'                             EOL ...
   ['b = ''' EOL ''';']                      EOL ...                             
    ];

  % expr with * to operator-late PFN
  postfix0  = [      % requires LD
    'g = e;'                                 EOL ...
    'e = t r*;'                              EOL ...
    'r = ''+'' t "+";'                       EOL ...
    'r = ''-'' t "-";'                       EOL ...
    't = f s*;'                              EOL ...
    's = ''*'' f "*";'                       EOL ...
    's = ''/'' f "/";'                       EOL ...
    'f = L;'                                 EOL ...
    'f = D;'                                 EOL ...
    'f = ''('' e '')'';'                     EOL ...
    ];

  % goal symbol  (defines phrase names g r f a l)
  Vg0   = [    %  Requires tag ULa
    'g = Lp*;'                               EOL ...
    'p = a;'                                 EOL ...
    ];
  
  % phrase names  (defines phrase names g r f a l)
  Vn0   = [    %  Requires tag aULul
    'g = r*;'                                EOL ...
    'r = L ''='' f '';'';'                   EOL ...
    'f = p*;'                                EOL ...
    'p = l;'                                 EOL ...
    'p = l ''*'';'                           EOL ...
    'p = l ''+'';'                           EOL ...
    'p = '''''' a '''''';'                   EOL ...
    'p = ''"'' a ''"'';'                     EOL ...
    ];

  % input symbols  (defines phrase names g r f a l)
  Vi0   = [    %  Requires tag aAlu
    'g = r*;'                                EOL ...
    'r = l ''='' f '';'';'                   EOL ...
    'f = p*;'                                EOL ...
    'p = l;'                                 EOL ...
    'p = l ''*'';'                           EOL ...
    'p = l ''+'';'                           EOL ...
    'p = '''''' A '''''';'                   EOL ...
    'p = ''"'' a ''"'';'                     EOL ...
    ];

  % output symbols  (defines phrase names g r f a l d)
  Vo0   = [    %  Requires tags aAlu
    'g = r*;'                                EOL ...
    'r = l ''='' f '';'';'                   EOL ...
    'f = p*;'                                EOL ...
    'p = l;'                                 EOL ...
    'p = l ''*'';'                           EOL ...
    'p = l ''+'';'                           EOL ...
    'p = '''''' a '''''';'                   EOL ...
    'p = ''"'' A ''"'';'                     EOL ...
    ];


  % transform away the + and *
  pretty1  = run(scan(pretty0), nostar1, 'ALU');
  pretty2  = nodup(run(scan(pretty0), nostar2, 'alu'), 8);
  pretty   = [pretty1 pretty2];
  
  % transform away the *
  post1    = run(scan(postfix0), nostar1, 'ALU');
  post2    = run(scan(postfix0), nostar2, 'alu');
  postfix  = [post1 post2];

  % transform away the *
  Vg1      = run(scan(Vg0), nostar1, 'ALU');
  Vg2      = run(scan(Vg0), nostar2, 'alu');
  Vg       = [Vg1 Vg2];
  Vn1      = run(scan(Vn0), nostar1, 'ALU');
  Vn2      = run(scan(Vn0), nostar2, 'alu');
  Vn       = [Vn1 Vn2];
  Vi1      = run(scan(Vi0), nostar1, 'ALU');
  Vi2      = run(scan(Vi0), nostar2, 'alu');
  Vi       = [Vi1 Vi2];
  Vo1      = run(scan(Vo0), nostar1, 'ALU');
  Vo2      = run(scan(Vo0), nostar2, 'alu');
  Vo       = [Vo1 Vo2];
  
  obj = public();
  %return;
  
  function otxt = run(itxt, gtxt, flags)         % run GEM
    if nargin == 2, flags = ''; end;
    otxt = iog4(itxt, gtxt, flags);              % call MEX
  end

  function otxt = scan(itxt)                     % deblank
    otxt = iog4(itxt, nowhite, 'A');             % call MEX
  end

  % Remove duplicates of length n from the input.
  % In a loop, 
  % 1. capture and remove the first n chars of input
  % 2. construct an IOG ignoring the capture and passing the rest
  % 3. apply the IOG, then repeat the process on the result.
  % 4. the concatenation of the captures is the answer.
  function final = nodup(dups, sz)
    assert(mod(numel(dups),sz)==0);
    final = '';
    subs = ['p=' char(zeros(1,sz)+'A') ';'];
    while numel(dups) > 0
      next = dups(1:sz);
      dups(1:sz) = [];
      gr = ['g=pg;g=;p=''' next ''';' subs];
      final = [final next];                      %#ok<AGROW>
      dups = run(dups, gr, 'A');                 % gr needs 'A'
    end
  end

  function o = public()
    o = struct;
    o.nostar1  = nostar1;
    o.nostar2  = nostar2;
    o.noplus   = noplus;

    o.postfix  = postfix;  % useable
    o.postfix0 = postfix0; % lookable
    o.pretty   = pretty;   % useable
    o.pretty0  = pretty0;  % lookable
    
    o.Vg0      = Vg0;      % goal symbol    (lookable)
    o.Vg       = Vg;       % goal symbol    (useable)
    o.Vn0      = Vn0;      % phrase names   (lookable)
    o.Vn       = Vn;       % phrase names   (useable)
    o.Vi0      = Vi0;      % input symbols  (lookable)
    o.Vi       = Vi;       % input symbols  (useable)
    o.Vo0      = Vo0;      % output symbols (lookable)
    o.Vo       = Vo;       % output symbols (useable)
    o.nodup    = @nodup;   % like unix unique
    
    o.nowhite  = nowhite;
    o.scan     = @scan;
    o.run      = @run;
  end

end

