% FILE:    gem5.m
% PURPOSE: machine to execute self-describing grammars
% METHOD:  The machine iog4.c has been described before.
%  gem5 emphasizes access  to the hardware using MEX file runX86.c
% EXAMPLES:     
%   G = gem5();                          % constructor

% COPYRIGHT:   1988 W. M. McKeeman.  See license.txt.
% MODS:	       1988.06.02 -- mckeeman -- original
%              2009.05.04 -- mckeeman@dartmouth -- move to MEX

function obj = gem5()
  EOL = char(10);                        % ascii end-of-line
  
  %    ------- input grammars -----------------
  
  % de-whiter-er
  nowhite = [        % requires A
    'g=pg;'           ...
    'g=;'             ...
    'p='' '';'        ...  discard blank
    'p=''' EOL ''';'  ...  discard EOL
    'p=III;'          ...  keep '''
    'p=OOO;'          ...  keep """
    'p=''''''i;'      ...  expand 'input stuff'
    'p=''"''o;'       ...  expand "output stuff"
    'p=A;'            ...  pass everything else
    'i='''''';'       ...  comes after input stuff
    'i="''"A"''"i;'   ...  expand one  input char
    'o=''"'';'        ...  comes after output stuff
    'o="""A"""o;'     ...  expand one output char
    'I=''''''"''";'   ...  see ', send '
    'O=''"''""";'     ...  see ", send "
    ];
  
  % an IOG pretty printer
  % handles extra and missing whitespace, * and +
  % leaves multicharacter inputs and outputs unchanged
  pretty0 = [        % requires LUA
    'g = b* r*;'                             EOL ...
   ['r = L b* ''='' " =" f* b* '';'' b* ";' EOL '";'] EOL ...
    'f = b* " " p;'                          EOL ...
    'p = I I I;'                             EOL ... 
    'p = I i "''";'                          EOL ...
    'p = O O O;'                             EOL ...
    'p = O o """;'                           EOL ...
    'p = L b* ''*'' "*";'                    EOL ...
    'p = L b* ''+'' "+";'                    EOL ...
    'p = L;'                                 EOL ...
    'i = '''''';'                            EOL ...
    'i = A i;'                               EOL ... 
    'o = ''"'';'                             EOL ...
    'o = A o;'                               EOL ...
    'I = '''''' "''";'                       EOL ...
    'O = ''"'' """;'                         EOL ...
    'b = '' '';'                             EOL ...
   ['b = ''' EOL ''';']                      EOL ...                             
    ];
  
  invert = [          % requires A
    'g = p g;'                       EOL ...
    'g = ;'                          EOL ...
    'p = '''''' """ A '''''' """;'   EOL ...
    'p = ''"'' "''" A ''"'' "''";'   EOL ...
    'p = A;'                         EOL ...
    ];

  % make new rules to define a* (only lower)     
  % 'a*' goes to "A=aA;A=;";
  nostar1 = [         % requires ALU   
    'g = r g;'                               EOL ...
    'g =;'                                   EOL ...
    'r = l ''='' f '';'';'                   EOL ...
    'f = p f;'                               EOL ...
    'f =;'                                   EOL ...
    'p = '''''' a '''''';'                   EOL ...
    'p = ''"'' a ''"'';'                     EOL ...
    'p = s;'                                 EOL ...
    'p = l;'                                 EOL ...
    sprintf(sprintf('s=''%c*''"%%c=%c%%c;%%c=;";', ...
      repeat('a':'z', 2)),repeat('A':'Z', 3))        EOL ...
    ];

  % replace a* with A in a grammar (only lower) 
  % a* goes to A
  nostar2 = [        % requires ALU
    'g = r g;'                               EOL ...
    'g =;'                                   EOL ...
    'r = L ''='' "="  f '';'' ";" ;'         EOL ...
    'f = p f;'                               EOL ...
    'f =;'                                   EOL ...
    'p = I A I;'                             EOL ...
    'p = O A O;'                             EOL ...
    'p = s;'                                 EOL ...
    'p = L;'                                 EOL ...
    'I = '''''' "''";'                       EOL ...  see ', send '
    'O = ''"''  """;'                        EOL ...  see ", send "
    sprintf('s=''%c*''"%c";', shuffle('a':'z','A':'Z'))  EOL ...
    ];

  % replace a+ in a grammar (only lower) 
  % a+ goes to aa*
  noplus = [         % requires ALU
    'g = r g;'                               EOL ...
    'g =;'                                   EOL ...
    'r = L ''='' "="  f '';'' ";" ;'         EOL ...
    'f = p f;'                               EOL ...
    'f =;'                                   EOL ...
    'p = I A I;'                             EOL ...
    'p = O A O;'                             EOL ...
    'p = S;'                                 EOL ...
    'p = L;'                                 EOL ...
    'I = '''''' "''";'                       EOL ...  see ', send '
    'O = ''"''  """;'                        EOL ...  see ", send "
    sprintf('S=''%c+''"%c%c*";', repeat('a':'z',3))  EOL ...
    ];

  prolog = [                   ...
    'pushR EBP'            EOL ... push the base pointer on the stack
    'movRR EBP ESP'        EOL ... replace the base with the stack pointer
    'pushA'                EOL ... save all the general registers
    ];
  
  epilog = [                   ...
    'popA'                 EOL ... restore the general registers
    'xor EAX EAX'          EOL ... zero return code means success
    'leave'                EOL ... restore stack
    'ret'                  EOL ... restore program counter
    ];

  % a tiny Intel X86 assembler (blank sep, EOL forced)
  asm = [                   
      'p = i p;'                                      EOL ...
      'p =;'                                          EOL ...
      'i = ''leave''         e            "C9";'      EOL ...
      'i = ''movRR EBP ESP'' e            "89E5";'    EOL ...
      'i = ''popA''          e            "61";'      EOL ...
      'i = ''pushA''         e            "60";'      EOL ...
      'i = ''pushR EBP''     e            "55";'      EOL ...
      'i = ''ret''           e            "C3";'      EOL ...
      'i = ''xor EAX EAX''   e            "33C0";'    EOL ...
     ['e = ''' EOL ''';']                             EOL ...
    ];

  % a 4-register calculator
  % a = 3;
  % a += b;
  % MOD     = @(m)bitshift(bitand(uint8(m),3),6);
  % REG     = @(r)bitshift(bitand(uint8(r),7),3);
  % R_M     = @(rm)bitand(uint8(rm),7);
  % MOD_REG = @(rm,r)bitor(bitor(MOD(3), REG(r)), R_M(rm));
  rmRR = [...
    'aaC0', 'abC8', 'acD0', 'adD8',...
    'baC1', 'bbC9', 'bcD1', 'bdD9',...
    'caC2', 'cbCA', 'ccD2', 'cdDA',...
    'daC3', 'dbCB', 'dcD3', 'ddDB'];
  
  mrRR = [...
    'aaC0', 'abC1', 'acC2', 'adC3',...
    'baC8', 'bbC9', 'bcCA', 'bdCB',...
    'caD0', 'cbD1', 'ccD2', 'cdD3',...
    'daD8', 'dbD9', 'dcDA', 'ddDB'];
  
  calc = [                 % Requires D
    'c = "554889E5" p "C9C3";'                    EOL ...
    'p = i p;'                                    EOL ...
    'p =;'                                        EOL ...
    'i = a;'                                      EOL ...
    'i = x;'                                      EOL ...
    'a = ''a='' "B80" h "000000" '';'' ;'         EOL ...
    'a = ''b='' "B90" h "000000" '';'' ;'         EOL ...
    'a = ''c='' "BA0" h "000000" '';'' ;'         EOL ...
    'a = ''d='' "BB0" h "000000" '';'' ;'         EOL ...
    'h = D;'                                      EOL ...
    sprintf('x = ''%c+=%c;'' "01%c%c";\n',   rmRR),...
    sprintf('x = ''%c-=%c;'' "2B%c%c";\n',   mrRR),...
    sprintf('x = ''%c*=%c;'' "0FAF%c%c";\n', mrRR),...
    sprintf('x = ''%c|=%c;'' "0B%c%c";\n',   mrRR)... 
    ];
    
  % convert ascii integer into binary integer
  atoi = [
    'c = "5589E5" "33C0" "B90A000000" p "C9C3";'  EOL ... pro/epilog
    'p = "0FAFC1" i p;'                           EOL ... EAX*=ECX;
    'p =;'                                        EOL ... EAX=0
    'i = ''0'';'                                  EOL ... EAX+=0;
    'i = ''1'' "40";'                             EOL ... EAX+=1;
    'i = ''2'' "0502000000";'                     EOL ... EAX+=2;
    'i = ''3'' "0503000000";'                     EOL ... EAX+=3;
    'i = ''4'' "0504000000";'                     EOL ... EAX+=4;
    'i = ''5'' "0505000000";'                     EOL ... EAX+=5;
    'i = ''6'' "0506000000";'                     EOL ... EAX+=6;
    'i = ''7'' "0507000000";'                     EOL ... EAX+=7;
    'i = ''8'' "0508000000";'                     EOL ... EAX+=8;
    'i = ''9'' "0509000000";'                     EOL ... EAX+=9;
    ];

  % convert ascii HEX characters into binary
  char2hex = ['g=hg;' EOL 'g=;' EOL];
  v = 0;
  for i=['0':'9' 'A':'F']; 
    for j=['0':'9' 'A':'F']
      char2hex = [char2hex 'h=''' i j ''' "' v '";' EOL]; %#ok<AGROW>
      v = v + 1;
    end; 
  end;

  % transform away the *
  prettygrammar = run(scan(pretty0), nostar2, 'LUA');
  prettyrules   = run(scan(pretty0), nostar1, 'lua');
  pretty        = [prettygrammar prettyrules];
    
  obj = public();
  %return;
  
  function otxt = run(itxt, gtxt, flags)    % run
    if nargin == 2, flags = ''; end;
    otxt = iog4(itxt, gtxt, flags);         % call mex
  end

  function otxt = scan(itxt)                % deblank
    otxt = iog4(itxt, nowhite, 'A');
  end
    
  % Pack MATLAB chars into hexadecimal.
  % This is a case of using MATLAB to do something essential, which
  % apparently violates the stated challenge of bootstrapping everything.
  % GEM is actually capable of this transformation (see char2hex)
  % but MATLAB/C interface cannot transmit it.
  % The problem is that machine language has lots of zeros (ascii NULL)
  % which truncate the output.
  function res = val(hexdigit)
    if '0'<=hexdigit && hexdigit<='9', res = hexdigit-'0';
    else res = hexdigit - 'A' + 10;
    end
  end

  % Run hex string on underlying hardware.
  % Squeeze 2 MATLAB chars into each code byte.
  function rc = exe(hex)
    switch computer
      case {'PCWIN', 'GLNX86', 'MACI'}
      otherwise
        error('Intel x86 hardware required');
    end
    nbytes = numel(hex)/2;
    code = zeros(1,nbytes,'uint8');
    for pc = 1:nbytes                       % squeeze out leading zeros
      b1 = val(hex(2*pc-1));
      b2 = val(hex(2*pc));
      code(pc) = bitor(bitshift(b1,4),b2);
    end
    nonox(code);
    rc = runX86(code);
  end

  function o = public()
    o = struct;
    o.nostar1  = nostar1;
    o.nostar2  = nostar2;
    o.noplus   = noplus;
    o.pretty   = pretty;   % useable
    o.pretty0  = pretty0;  % lookable
    o.nowhite  = nowhite;
    o.char2hex = char2hex;
    o.invert   = invert;

    o.prolog   = prolog;
    o.epilog   = epilog;
    o.asm      = asm;
    o.calc     = calc;
    o.atoi     = atoi;
    
    o.scan     = @scan;
    o.run      = @run;
    o.exe      = @exe;
  end

end

