/* FILE:        iog0.c
 * PURPOSE:     Machine to execute self-describing grammars
 * COPYRIGHT:   1988 W. M. McKeeman.  You may do anything you like
 *              with this file except remove or alter this notice.
 * METHOD:      Skating on thin ice -- this code is always on the edge
 *              of catastrophic failure.  See iog1.c for an implementation
 *              that pays attention to reliability.
 * MODS:        2008.06.11 -- mckeeman -- ported to c89 MEX
 *              2009.01.16 -- mckeeman -- removed character classes
 *              2009.04.08 -- luhe     -- stripped to basics
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "mex.h"

#define DATALIM  10000
#define STACKLIM 10000

#define C(c) case c:
#define ALPHA UPPER LOWER
#define LOWER \
C('a')C('b')C('c')C('d')C('e')C('f')C('g')C('h')C('i')C('j')C('k')C('l')C('m')\
C('n')C('o')C('p')C('q')C('r')C('s')C('t')C('u')C('v')C('w')C('x')C('y')C('z')
#define UPPER \
C('A')C('B')C('C')C('D')C('E')C('F')C('G')C('H')C('I')C('J')C('K')C('L')C('M')\
C('N')C('O')C('P')C('Q')C('R')C('S')C('T')C('U')C('V')C('W')C('X')C('Y')C('Z')

#define SEARCH 1
#define BACK   2
#define PARSE  3

#define psize (ps - &parsestack[0])
#define bsize (bs - &backstack[0])

static char  code[DATALIM];
static char  input[DATALIM];
static char  output[DATALIM];
static char *parsestack[STACKLIM];
static char *backstack[STACKLIM];
static char  ini[7];

static char *p0, *p, **ps, **bs;                /* program pointers */
static int   pN;                                /* end of program */
static char *i0, *i;                            /* input pointers */
static char *o0, *o;                            /* output pointers */
static int   mode;                              /* PARSE/BACK/SEARCH */


static void
error(const char *msg) {
  char errmsg[100];
  sprintf(errmsg, 
     "gem error.  input: %s\ngem: %s p=%d  *p='%c'  i=%d  o=%d\n",
     input, msg, p-p0, *p==0?'#':*p, i-i0, o-o0);
  mexErrMsgTxt(errmsg);
}

static void
clear(char *b, int n) {
  int idx;
  for (idx=0;idx<n;idx++) b[idx]=0;
}

static void
pregem(const char *gr, const char *in) {
  char goal;
  int initlen;
  
  clear(code, sizeof(code));                /* make it all zeros */
  strcpy(code, "=G'X';");                   /* zero-th rule */
  initlen = strlen(code);                   /* size of init string */
  p = p0 = code+initlen-1;                  /* point at ; */
  goal = gr[0];                             /* user goal symbol */
  code[1] = goal;                           /* overwrite G */
  if (initlen+strlen(gr)>=sizeof(code)) error("input grammar too long");
  strcat(code, gr);                         /* add rest of user code */
  pN = strlen(code)-initlen+1;              /* end of code */
  code[3] = 0;                              /* overwrite X */
  
  clear(input, sizeof(input));              /* make it all zeros */
  i = i0 = input;                           /* start of "input" */
  if (strlen(in)>=sizeof(input)) error("input text too long");
  strcpy(input, in);                        /* null terminated */
  
  clear(output, sizeof(output));            /* make it all zeros */
  o = o0 = output;                          /* start of output */
  o--;                                      /* empty "output" */
  
  ps = &parsestack[0];                      /* code ptr stack empty */
  *ps = code;                               /* pretend previous progress */
  (*ps)++;                                  /* point at G */
  **ps = goal;                              /* replace G with goal symbol */    
  bs = &backstack[0];
  bs--;                                     /* empty "parse" */

  mode = SEARCH;                            /* looking at ; in code[0] */
 
}

# include "iog0.h"

static char *
postgem(void) {
  o++; *o = '\0';                           /* null terminate "output" */
  return output;
}

/* ************************************************************** */
void
mexFunction(
  int nlhs, mxArray *plhs[],
  int nrhs, const mxArray *prhs[]) {
  char *o, *g, *i, *f;
  
  /* check inputs */
  mxAssert(nrhs == 2, "bad call: use gem(i,g)");    
  mxAssert(nlhs<=1, "bad call: use o = gem(i,g)");

  mxAssert(mxIsChar(prhs[0]), "bad call: requires char 1st arg");    
  i = mxArrayToString(prhs[0]);             /* the input */
  mxAssert(mxIsChar(prhs[1]), "bad call: requires char 2nd arg");   
  g = mxArrayToString(prhs[1]);             /* the grammar */
 
  
  pregem(g, i);                          /* set up globals */
  gem();                                 /* translate */
  o = postgem();                         /* report result */
  
  mxFree(g);
  mxFree(i);
  
  plhs[0] = mxCreateString(o);
}

