/* FILE:        iog2.c
 * PURPOSE:     Machine to execute self-describing grammars
 * COPYRIGHT:   1988 W. M. McKeeman.  You may do anything you like
 *              with this file except remove or alter this notice.
 * METHOD:      extend log1.c, add whitespace processed build-in          
 * MODS:        2008.06.11 -- mckeeman -- ported to c89 MEX
 *              2009.01.16 -- mckeeman -- removed character classes
 *              2009.04.10 -- luhe     -- add whitespace processed build-in
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "mex.h"

#define C(c) case c:
#define ALPHA UPPER LOWER
#define LOWER \
C('a')C('b')C('c')C('d')C('e')C('f')C('g')C('h')C('i')C('j')C('k')C('l')C('m')\
C('n')C('o')C('p')C('q')C('r')C('s')C('t')C('u')C('v')C('w')C('x')C('y')C('z')
#define UPPER \
C('A')C('B')C('C')C('D')C('E')C('F')C('G')C('H')C('I')C('J')C('K')C('L')C('M')\
C('N')C('O')C('P')C('Q')C('R')C('S')C('T')C('U')C('V')C('W')C('X')C('Y')C('Z')

#define SEARCH 1
#define BACK   2
#define PARSE  3

/* bounds and type checking macros */
#define DATALIM  10000
#define STACKLIM 1000
#define SIZE(A) (int)(sizeof(A)/sizeof(*(A)))
#define psize (int)(ps - &parsestack[0])
#define bsize (int)(bs - &backstack[0])
#define CHECKPARSE() if (psize < 0 || psize >= SIZE(parsestack)) error("PARSE stack overflow");
#define CHECKBACK()  if (bsize < 0 || bsize >= SIZE(backstack))  error("BACK stack overflow");
#define CHECKNAME(n) if (!isalpha(n)) error("rule name must be alpha")
#define CHECKCODE()  if (p-p0 >= pN) error("ran beyond code")
#define CHECKOUT()   if (o<o0 || o-o0 >= sizeof(output)) error("too much output")

static char  code[DATALIM];
static char  input[DATALIM];
static char  output[DATALIM];
static char *parsestack[STACKLIM];
static char *backstack[DATALIM];
static char  ini[7];

static char *p0, *p, **ps, **bs;                /* program pointers */
static int   pN;                                /* end of program */
static char *i0, *i;                            /* input pointers */
static char *o0, *o;                            /* output pointers */
static int   mode;                              /* PARSE/BACK/SEARCH */
static int   traceit;

/* ------------------- helper functions ------------------- */
static void 
TRACE(void) {
  if (psize>=0)
  printf("%s  stk:%d **ps:%c  p:%d *p:%c  i:%d *i:%c  o:%d *o:%c\n",
  mode == PARSE ? " PARSE" : mode == BACK ? "  BACK" : "SEARCH", 
  psize, **ps,
  p-p0, p-p0>=0&&*p!=0?*p:'#', 
  i-i0, i-i0>=0&&*i!=0?*i:'#',
  o-o0, o-o0>=0&&*o!=0?*o:'#');
}

static void
error(const char *msg) {
  char errmsg[DATALIM+100];
  sprintf(errmsg, 
     "gem error.  input: %s\ngem: %s p=%d  *p='%c'  i=%d  o=%d\n",
     input, msg, p-p0, *p==0?'#':*p, i-i0, o-o0);
  mexErrMsgTxt(errmsg);
}

static void
clear(char *b, int n) {
  int idx;
  for (idx=0;idx<n;idx++) b[idx]=0;
}

static void
pregem(const char *gr, const char *in, const char *flag) {
  char goal;
  int initlen;
  int idx, grlen;
  clear(code, sizeof(code));                /* make it all zeros */
  strcpy(code, "=G'X';");                   /* zero-th rule */
  initlen = strlen(code);                   /* size of init string */
  p = p0 = code+initlen-1;                  /* point at ; */
  grlen = strlen(gr);
  for (idx=0; idx<grlen; idx++) {           /* skip initial blanks */
    if (gr[idx]!=' ' && gr[idx]!='\n') break;
  }
  goal = gr[idx];                           /* user goal symbol */
  code[1] = goal;                           /* overwrite G */
  if (initlen+strlen(gr)>=sizeof(code)) error("input grammar too long");
  strcat(code, gr);                         /* add rest of user code */
  pN = strlen(code)-initlen+1;              /* end of code */
  code[3] = 0;                              /* overwrite X */
  
  clear(input, sizeof(input));              /* make it all zeros */
  i = i0 = input;                           /* start of "input" */
  if (strlen(in)>=sizeof(input)) error("input text too long");
  strcpy(input, in);                        /* null terminated */
  
  clear(output, sizeof(output));            /* make it all zeros */
  o = o0 = output;                          /* start of output */
  o--;                                      /* empty "output" */
  
  ps = &parsestack[0];                      /* code ptr stack empty */
  *ps = code;                               /* pretend previous progress */
  (*ps)++;                                  /* point at G */
  **ps = goal;                              /* replace G with goal symbol */
  bs = &backstack[0];
  bs--;                                     /* empty "parse" */

  mode = SEARCH;                            /* looking at ; in code[0] */
  traceit = strcmp(flag, "T") == 0;
}

# include "iog2.h"

static char *
postgem(void) {
  o++; *o = '\0';                           /* null terminate "output" */
  return output;
}

/* ************************************************************** */
void
mexFunction(
  int nlhs, mxArray *plhs[],
  int nrhs, const mxArray *prhs[]
) {
  char *o, *g, *i, *f;
  
  /* check inputs */
  mxAssert(nrhs == 2 || nrhs == 3, "bad call: use gem(i,g) or gem(i,g,f)");    
  mxAssert(nlhs<=1, "bad call: use o = gem(i,g)");

  mxAssert(mxIsChar(prhs[0]), "bad call: requires char 1st arg");    
  i = mxArrayToString(prhs[0]);             /* the input */
  mxAssert(mxIsChar(prhs[1]), "bad call: requires char 2nd arg");   
  g = mxArrayToString(prhs[1]);             /* the grammar */
  if (nrhs == 3) {
    mxAssert(mxIsChar(prhs[2]), "bad call: requires char flag");   
    f = mxArrayToString(prhs[2]);           /* a flag */
  } else {
    f = "";                                 /* no flag */
  }
  
  pregem(g, i, f);                          /* set up globals */
  gem();                                    /* translate */
  o = postgem();                            /* report result */
  
  mxFree(g);
  mxFree(i);
  if (nrhs == 3) mxFree(f);
  
  plhs[0] = mxCreateString(o);
}

