/* FILE:        iog6.c
 * PURPOSE:     Machine to execute self-describing grammars
 * COPYRIGHT:   1988 W. M. McKeeman.  You may do anything you like
 *              with this file except remove or alter this notice.
 * METHOD:      extend log4.c, multicharacter phase name <name> built in          
 * MODS:        2008.06.11 -- mckeeman -- ported to c89 MEX
 *              2009.01.16 -- mckeeman -- removed character classes
 *              2009.05.04 -- luhe     -- add multicharacter phase name 
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "mex.h"

#define C(c) case c:
#define ALPHA UPPER LOWER
#define LOWER \
      C('b')C('c')      C('e')C('f')C('g')C('h')C('i')C('j')C('k')      C('m')\
C('n')C('o')C('p')C('q')C('r')C('s')C('t')C('u')C('v')C('w')C('x')C('y')C('z')
#define UPPER \
      C('B')C('C')      C('E')C('F')C('G')C('H')C('I')C('J')C('K')      C('M')\
C('N')C('O')C('P')C('Q')C('R')C('S')C('T')C('U')C('V')C('W')C('X')C('Y')C('Z')

#define SEARCH 1
#define BACK   2
#define PARSE  3


/* bounds and type checking macros */
#define DATALIM  10000
#define STACKLIM 1000
#define SIZE(A) (int)(sizeof(A)/sizeof(*(A)))
#define psize (int)(ps - &parsestack[0])
#define bsize (int)(bs - &backstack[0])
#define CHECKPARSE() if (psize < 0 || psize >= SIZE(parsestack)) error("PARSE stack overflow")
#define CHECKBACK()  if (bsize < 0 || bsize >= SIZE(backstack))  error("BACK stack overflow")


#define CHECKNAME(n) if (!isalpha(n)) error("single-character rule name must be alpha")
#define CHECKMULTIName(n) if(n!='_' && !isalpha(n) && !isdigit(n)) error("unexpected multi-character phrase name")
#define CHECKALPHA(n) if(!isalpha(n)) error("multi-character phrase name must start with alpha")


#define CHECKCODE()  if (p-p0 >= pN) error("ran beyond code")
#define CHECKOUT()   if (o<o0 || o-o0 >= sizeof(output)) error("too much output")
#define CHECKIN()    if (i<i0 || i-i0 >= sizeof(input)) error("too much input")
#define SEARCHANDRECUR                                            \
  mode = SEARCH;                /* not builtin L: search on */    \
  ps++; CHECKPARSE(); *ps = p;  /* save return address */         \
  p = p0;                       /* start SEARCH at beginning */

#define UNCALL                                                    \
  ps++; CHECKPARSE(); *ps = p;  /* save return address */         \
  p = *bs; bs--;                /* forget backtrack point */             

static char  code[DATALIM];
static char  input[DATALIM];
static char  output[DATALIM];
static char *parsestack[STACKLIM];
static char *backstack[DATALIM];
static char  ini[7];

static char *p0, *p, **ps, **bs;                /* program pointers */
static int   pN;                                /* end of program */
static int   iN;                                /* end of input */
static char *i0, *i;                            /* input pointers */
static char *o0, *o;                            /* output pointers */
static int   mode;                              /* PARSE/BACK/SEARCH */

static int   traceit;                           /* flags */
static int   digitCFG, lowerCFG, upperCFG, asciiCFG;
static int   digitIOG, lowerIOG, upperIOG, asciiIOG; 



/* ------------------- helper functions ------------------- */
static char
vis(char c) {
  return c < ' ' ? '.' : c;
}
          
static void
TRACE(void) {
  char *tmode;
  int j, tlim, clen=30, ilen=16, olen=16;
  tmode = mode == PARSE ? " PARSE" : mode == BACK ? "  BACK" : "SEARCH";
  printf("%s ", tmode); 
  for (j=-clen/2;j<clen/2;j++) {
    if (j==0) printf("|");
    if (p+j<=p0) printf(" ");
    else if (p+j>p0+pN) printf(" ");
    else printf("%c", vis(*(p+j)));
  }

  printf("  ");   /* begin printing input */
  for (j=0;j<ilen;j++) {  /* ilen chrs */
    if (i+j<i0+iN) printf("%c", vis(*(i+j)));
    else printf(" ");
  }
  
  printf("  ");   /* begin printing output */
  for (j=olen-1;j>=0;j--) {  /* olen chrs */
    if (o-j>=o0) printf("%c", vis(*(o-j)));
    else printf(" ");
  }
  printf("\n");
}

static void
error(const char *msg) {
  char errmsg[DATALIM+100];
  sprintf(errmsg, 
     "gem error.  input:\n%s\ngem: %s p=%d  *p='%c'  i=%d  o=%d\n",
     input, msg, p-p0, *p==0?'#':*p, i-i0, o-o0);
  mexErrMsgTxt(errmsg);
}

static void
clear(char *b, int n) {
  int idx;
  for (idx=0;idx<n;idx++) b[idx]=0;
}

static void
pregem(const char *gr, const char *in, const char* flag) { 
  int idx, gdx, cdx, nullspot, semispot, grlen;
  int debug;
  clear(input, sizeof(input));                   /* make it all zeros */
  i = i0 = input;                                /* start of "input" */
  iN = strlen(in);
  if (strlen(in)>=sizeof(input)) error("input text too long");
  strcpy(input, in);                             /* null terminated */
  
  clear(output, sizeof(output));                 /* make it all zeros */
  o = o0 = output;                               /* start of output */
  o--;                                           /* empty "output" */

  grlen = strlen(gr);
  clear(code, sizeof(code));                     /* make it all zeros */
  for (idx=0; idx<grlen; idx++) {                /* skip initial blanks */
    if (gr[idx]!=' ' && gr[idx]!='\n') break;    /* first nonblank */
  }  
  cdx = 0;                                       /* init sequence */
  code[cdx++] = '=';                             /* = */
  if (gr[idx] == '<') {                          /* <g> */
    for (gdx=idx; idx<grlen; gdx++) {
      code[cdx++] = gr[gdx];                     /* transfer chars */
      if (gr[gdx] == '>') break;                 /* > was moved */
    }
  } else {                                       /* g */
    code[cdx++] = gr[idx];
  }
  
  code[cdx++] = '\'';                            /* 'X'; */
  nullspot = cdx;                                /* X gets zapped later */
  code[cdx++] = 'X';                             /* place for EOF */
  code[cdx++] = '\'';
  semispot = cdx;                                /* to find ; again */
  code[cdx++] = ';';
  if (cdx+grlen >= sizeof(code)) error("input grammar too long");
  strcat(code, gr);                              /* add user code */ 
  

  pN = strlen(gr)+1;                             /* end of code */
  p = p0 = code+semispot;                        /* point at ; */
  code[nullspot] = 0;                            /* EOF code */
  
    
  ps = &parsestack[0];                           /* code ptr stack empty */
  *ps = code;                                    /* pretend previous progress */
  (*ps)++;                                       /* point at G or < */
  bs = &backstack[0];
  bs--;                                          /* empty "parse" */

  mode = SEARCH;                                 /* looking at ; in code */
  
  traceit=0;                                     /* unset the flags */
  digitIOG=0; digitCFG=0; lowerIOG=0; lowerCFG=0; 
  upperIOG=0; upperCFG=0; asciiIOG=0; asciiCFG=0;
  
  for (idx=0; idx<(int)strlen(flag); idx++) {    /* set the flags */
    switch (flag[idx]){
      case 'D': digitIOG=1; break;
      case 'd': digitCFG=1; break;
      case 'L': lowerIOG=1; break;
      case 'l': lowerCFG=1; break;
      case 'U': upperIOG=1; break;
      case 'u': upperCFG=1; break; 
      case 'A': asciiIOG=1; break;
      case 'a': asciiCFG=1; break;
      case 'T': traceit =1; break;
      default:  error("Unexpected flag");
    }
  }  
  if (traceit) {
    printf("Tracing: three columns, each a moving window.\n");
    printf("  The | marks the current point of execution.\n");
    printf("  The head of the unconsumed input is shown.\n");
    printf("  The most recent output is shown.\n");
    printf("  Non-printable characters (including EOL) are shown as .\n");
    printf(" mode                code                      input        output\n");
  }
}

# include "iog6.h"                                /* the IOG execution */

static char *
postgem(void) {
  o++; *o = '\0';                           /* null terminate "output" */
  return output;
}

/* ************************************************************** */
void
mexFunction(
  int nlhs, mxArray *plhs[],
  int nrhs, const mxArray *prhs[]
) {
   char *o, *g, *i, *f;
  
  mxAssert(mxIsChar(prhs[0]), "bad call: requires char 1st arg");    
  i = mxArrayToString(prhs[0]);             /* the input */
  mxAssert(mxIsChar(prhs[1]), "bad call: requires char 2nd arg");   
  g = mxArrayToString(prhs[1]);             /* the grammar */
  if (nrhs == 3) {
    mxAssert(mxIsChar(prhs[2]), "bad call: requires char flag");   
    f = mxArrayToString(prhs[2]);           /* a flag */
  } else {
    f = "";                                 /* no flag */
  }
  
  pregem(g, i, f);                          /* set up globals */
  gem();                                    /* translate */
  o = postgem();                            /* report result */
  
  mxFree(g);
  mxFree(i);
  if (nrhs == 3) mxFree(f);
  
  plhs[0] = mxCreateString(o);
}

